/*
** Module   : TempTableBuilder.java
** Abstract : Implementation of the dynamic features for constructing temp-tables  
*              (attributes, methods and constructors).
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 OM  20130129 Created initial version.
** 002 VMN 20130217 Added support for handle-based attributes:
**                  ERROR (KW_ERROR)
** 003 CA  20130221 Added ADM-DATA, UNIQUE-ID, HandleChain and Deletable support.
** 004 CA  20130311 Added missing ADD-NEW-FIELD and ADD-NEW-INDEX API versions, using wrapper 
**                  types. Note that the conversion will use only the APIs with the wrapper type
**                  parameters; the other APIs (using non-wrapper parameters) are removed.
** 005 CS  20130322 Made the class non-abstract and added missing method stubs. 
** 006 CA  20130603 resourceDelete must return a boolean value.
** 006 SVL 20130624 Some methods were implemented.
** 007 CA  20130918 Added widget-pool support. The pool name must be validated before anything
**                  else, and a null name means the default pool is used.
** 008 SVL 20130822 Some methods and attributes were implemented.
** 009 CA  20131021 Adds DYNAMIC attribute support. Adds "asHandle" API to build a TempTable from
**                  a buffer associated with a temp-table.
** 010 CA  20131013 Added implicit resource deletion for static temp-table cases (the static
**                  resource gets deleted ONLY when its INSTANTIATING-PROCEDURE gets deleted).
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.util.*;
import org.hibernate.*;
import org.hibernate.mapping.PersistentClass;
import java.util.*;
import static com.goldencode.p2j.persist.DynamicTablesHelper.*;

/**
 * Implementation of buffer of a dynamic TEMP-TABLE and the associated 
 * attributes and methods.
 * This {@link TempTable} interface is needed by {@link CommonHandle}
 * in order to store the handle to a the table.
 * 
 * @author om
 */
public class TempTableBuilder
extends HandleChain
implements TempTable,
           Deletable
{
   /** datetime-tz 4GL type name. */
   private final static String DATETIMETZ_TYPE_NAME = "datetime-tz";

   /** Special characters (not letters or digits) allowed in dynamic table names. */
   private final static Set<Character> allowedSpecialChars = new HashSet<Character>();

   /**
    * Used during data type name validation. Map that stores the minimal possible length of a data
    * type name abbreviation. E.g. instead "integer" you can specify "integ", "int" or "i" because
    * the minimal length for this type is 1.
    */
   private final static Map<String, Integer> datatypesMatchLen =
                                                new LinkedHashMap<String, Integer>();

   /**
    * <code>true</code> if changes made to the records the temporary table can be undone.
    */
   private boolean undoable = false;

   /** Table fields keyed by their <i>normalized</i> names. */
   private final Map<String, P2JField> fields = new LinkedHashMap<String, P2JField>();

   /** Table indexes keyed by their <i>normalized</i> names. */
   private final Map<String, P2JIndex> indexes = new LinkedHashMap<String, P2JIndex>();

   /** Generated DMO buffer class. */
   private Class dmoIface = null;

   /** Handle to default buffer created for the temporary table. */
   private handle defaultBufferHandle = null;

   /** Corresponds ERROR-STRING attribute. **/
   private final character errorString = new character();

   /** Corresponds UNIQUE-ID attribute. **/
   private final integer uniqueId = new integer();

   /** Corresponds ERROR attribute. **/
   private final logical error = new logical();

   /** Corresponds ADM-DATA attribute. **/
   private final character admData = new character();

   /** Persistence class generated for this temporary table. **/
   private PersistentClass persistentClass = null;

   /** When <code>true</code>, marks this resource as being dynamically created. */
   private boolean dynamic = false;
   
   /**
    * Flag that indicates that an ADD-* function was called for the table (CLEAR function resets
    * this flag). The value of this flag is used in CREATE-LIKE.
    */
   private boolean addFunctionCalled = false;

   static
   {
      // should go first, see validateDatatype
      datatypesMatchLen.put(DATETIMETZ_TYPE_NAME, 11);
      datatypesMatchLen.put(ParmType.DT.toString(), 8);
      datatypesMatchLen.put(ParmType.INT64.toString(), 5);

      datatypesMatchLen.put(ParmType.INT.toString(), 1);
      datatypesMatchLen.put(ParmType.CHAR.toString(), 1);
      datatypesMatchLen.put(ParmType.DATE.toString(), 2);
      datatypesMatchLen.put(ParmType.DEC.toString(), 2);
      datatypesMatchLen.put(ParmType.HANDLE.toString(), 6);
      datatypesMatchLen.put(ParmType.LOG.toString(), 1);
      datatypesMatchLen.put(ParmType.RAW.toString(), 2);
      datatypesMatchLen.put(ParmType.RECID.toString(), 2);
      datatypesMatchLen.put(ParmType.ROWID.toString(), 3);

      allowedSpecialChars.addAll(Arrays.asList('-', '#', '$', '%', '&', '_'));
   }

   /**
    * Create a new TEMP-TABLE resource.
    * 
    * @param    dynamic
    *           flag markin this resource as dynamic, when set to <code>true</code>.
    */
   public TempTableBuilder(boolean dynamic)
   {
      this.dynamic = dynamic;
   }
   
   /**
    * Dynamically creates a temporary table at runtime and assign it to a 
    * handle. The table is empty and will be populated with fields, indexes
    * and relations subsequent statements.
    * This method is the P2J equivalent of <code>CREATE TEMP-TABLE</code> 
    * statement of Progress 4GL.
    * 
    * @param   tth
    *          The handle to assign the newly created temporary table to.
    * @param   widgetPool
    *          The name of the widget pool that contains the temp-table. <code>null</code> if the
    *          unnamed pool must be used.
    */
   public static void create(handle tth, String widgetPool)
   {
      create(tth, (widgetPool == null ? null : new character(widgetPool)));
   }

   /**
    * Dynamically creates a temporary table at runtime and assign it to a 
    * handle. The table is empty and will be populated with fields, indexes
    * and relations subsequent statements.
    * This method is the P2J equivalent of <code>CREATE TEMP-TABLE</code> 
    * statement of Progress 4GL.
    * 
    * @param   tth
    *          The handle to assign the newly created temporary table to.
    * @param   widgetPool
    *          The name of the widget pool that contains the temp-table. <code>null</code> if the
    *          unnamed pool must be used.
    */
   public static void create(handle tth, character widgetPool)
   {
      // validate widget pool before anything else!!!
      if (!WidgetPool.validWidgetPool(widgetPool))
      {
         return;
      }

      TempTableBuilder builder = new TempTableBuilder(true);
      builder.uniqueId.assign(UniqueIdGenerator.getUniqueId(UniqueIdGenerator.IdKind.TEMP_TABLE));
      tth.assign(builder);
      WidgetPool.addResource(builder, widgetPool);
   }

   /**
    * Dynamically creates a temporary table at runtime and assign it to a 
    * handle. The table is empty and will be populated with fields, indexes
    * and relations subsequent statements.
    * This method is the P2J equivalent of <code>CREATE TEMP-TABLE</code> 
    * statement of Progress 4GL.
    *
    * @param   tth
    *          The handle to assign the newly created temporary table to.
    */
   public static void create(handle tth)
   {
      create(tth, (character) null);
   }

   /**
    * Get the TEMP-TABLE handle associated with this temporary buffer.
    *
    * @param    buf
    *           A temporary buffer for which the associated TEMP-TABLE resource is needed.
    *
    * @return   See above.
    */
   public static handle asHandle(Temporary buf)
   {
      // TODO: set the DYNAMIC attribute depending on how this temp-table was built
      
      return new handle();
   }

   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the definitions.
    * @param   indexName
    *          The name of a single index which is copied from the source table.
    *
    * @return  <code>true</code> on success.
    */
   public logical createTableLike(handle hbuf, String indexName)
   {
      return createTableLike(hbuf, new character(indexName));
   }

   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the definitions.
    * @param   indexName
    *          The name of a single index which is copied from the source table.
    *
    * @return  <code>true</code> on success.
    */
   public logical createTableLike(handle hbuf, character indexName)
   {
      return createTableLikeImpl(hbuf, indexName, true);
   }

   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   tableName
    *          The name of a table to create the copy of.
    * @param   indexName
    *          The name of a single index which is copied from the source table.
    *
    * @return  <code>true</code> on success.
    */
   public logical createTableLike(String tableName, String indexName)
   {
      return createTableLike(new character(tableName), new character(indexName));
   }


   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   tableName
    *          The name of a table to create the copy of.
    * @param   indexName
    *          The name of a single index which is copied from the source table.
    *
    * @return  <code>true</code> on success.
    */
   public logical createTableLike(character tableName, character indexName)
   {
      return createTableLikeImpl(tableName, indexName, true);
   }

   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the definitions.
    *
    * @return  <code>true</code> on success.
    */
   public logical createTableLike(handle hbuf)
   {
      return createTableLikeImpl(hbuf, new character(), false);
   }

   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   tableName
    *          The name of a table to create the copy of.
    *
    * @return  <code>true</code> on success.
    */
   public logical createTableLike(String tableName)
   {
      return createTableLike(new character(tableName));
   }

   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   tableName
    *          The name of a table to create the copy of.
    *
    * @return  <code>true</code> on success.
    */
   public logical createTableLike(character tableName)
   {
      return createTableLikeImpl(tableName, new character(), false);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    *
    * @return  <code>true</code> on success.
    */
   public logical addNewField(character name, character type)
   {
      return addNewField(name, type, null, null, null, null, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    *
    * @return  <code>true</code> on success.
    */
   public logical addNewField(character name, character type, integer extent)
   {
      return addNewField(name, type, extent, null, null, null, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    *          
    * @return  <code>true</code> on success.
    */
   public logical addNewField(character name, character type, integer extent, character format)
   {
      return addNewField(name, type, extent, format, null, null, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    *          
    * @return  <code>true</code> on success.
    */
   public logical addNewField(character    name,
                              character    type,
                              integer      extent,
                              character    format,
                              BaseDataType initial)
   {
      return addNewField(name, type, extent, format, initial, null, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    * @param   label
    *          The label of the defined field. If null or unknown the name 
    *          parameter will be used.
    *          
    * @return  <code>true</code> on success.
    */
   public logical addNewField(character    name,
                              character    type,
                              integer      extent,
                              character    format,
                              BaseDataType initial,
                              character    label)
   {
      return addNewField(name, type, extent, format, initial, label, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * <p>
    * When a parameter is set to null, it means it was not specified by the business logic.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    * @param   label
    *          The label of the defined field. If null or unknown the name 
    *          parameter will be used.
    * @param   columnLabel
    *          The label of the column associated with the defined field
    *          
    * @return  <code>true</code> on success.
    */
   public logical addNewField(character    name,
                              character    type,
                              integer      extent,
                              character    format,
                              BaseDataType initial,
                              character    label,
                              character    columnLabel)
   {
      addFunctionCalled = true;

      if (_prepared())
      {
         displayPrepared();
         return new logical(false);
      }

      String fieldName = name.getValue();
      String typeName = type.getValue();

      if (fieldName == null || typeName == null)
      {
         displayUnableToEvaluate();
         return new logical(false);
      }

      if (hasField(fieldName))
      {
         // we have the field with the same name
         return new logical(false);
      }

      // validate data type
      ParmType parmType = validateDatatype(typeName);
      if (parmType == null)
         return new logical(false);

      String formatStr = format != null ? format.getValue() : null;
      if (formatStr != null)
      {
         try
         {
            // validate format
            switch (parmType)
            {
               case INT:
               case DEC:
               case INT64:
               case HANDLE:
               case RECID:
                  new NumberFormat(formatStr);
                  break;

               case DATE:
                  new date().toString(formatStr);
                  break;

               // TODO handle datetime[-tz] format validation
            }
         }
         catch (DisplayFormatParsingException e)
         {
            return new logical(false);
         }
      }

      String labelStr = label != null ? label.getValue() : null;
      if (labelStr == null)
         labelStr = fieldName;

      String columnLabelStr = columnLabel != null ? columnLabel.getValue() : null;
      if (columnLabelStr == null)
         columnLabelStr = labelStr;

      P2JField field = new P2JField(fieldName,
                                    parmType,
                                    extent != null ? extent.getValue() : 0,
                                    formatStr,
                                    initial,
                                    labelStr,
                                    columnLabelStr);
      addField(field);
      return new logical(true);
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * 
    * @return  <code>true</code> on success.
    */
   public logical addNewIndex(character name)
   {
      return addNewIndex(name, null, null, null);
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * 
    * @return  <code>true</code> on success.
    */
   public logical addNewIndex(character name, logical unique)
   {
      return addNewIndex(name, unique, null, null);
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * @param   primary
    *          TRUE if this is the primary index.
    * 
    * @return  <code>true</code> on success.
    */
   public logical addNewIndex(character name, logical unique, logical primary)
   {
      return addNewIndex(name, unique, primary, null);
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    * <p>
    * If a parameter is null, it means it was not set.
    * 
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * @param   primary
    *          TRUE if this is the primary index.
    * @param   word
    *          TRUE if this is a word index.
    * 
    * @return  <code>true</code> on success.
    */
   public logical addNewIndex(character name, logical unique, logical primary, logical word)
   {
      addFunctionCalled = true;

      if (_prepared())
      {
         displayPrepared();
         return new logical(false);
      }

      String indexName = name.getValue();
      if(indexName == null)
      {
         displayUnableToEvaluate();
         return new logical(false);
      }

      if (hasIndex(indexName))
      {
         displayDuplicateIndex(indexName);
         return new logical(false);
      }

      P2JIndex index = new P2JIndex(null,        // table name is set later in setPersistentClass
                                    indexName,
                                    unique != null && unique.getValue(),
                                    primary != null && primary.getValue(),
                                    word != null && word.getValue());
      addIndex(index);
      return new logical(true);
   }

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Database field name from which to copy the field. The table 
    *          name must be qualified with the database name.
    * 
    * @return  <code>true</code> on success.
    */
   public logical addFieldLike(String name, String source)
   {
      return addFieldLike(new character(name), new character(source));
   }

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Database field name from which to copy the field. The table 
    *          name must be qualified with the database name.
    * 
    * @return  <code>true</code> on success.
    */
   public logical addFieldLike(character name, character source)
   {
      addFunctionCalled = true;

      if (_prepared())
      {
         displayPrepared();
         return new logical(false);
      }

      String targetName = name.getValue();
      String fullSource = source.getValue();
      if (targetName == null || fullSource == null)
      {
         displayUnableToEvaluate();
         return new logical(false);
      }

      if (hasField(targetName))
      {
         // we have the field with the same name
         return new logical(false);
      }

      int pos = fullSource.lastIndexOf('.');
      if (pos < 0)
      {
         ErrorManager.displayError(9053,
                                   "Field-name in ADD-LIKE-FIELD must have a table qualifier",
                                   false);
         return new logical(false);
      }

      String srcFieldName = fullSource.substring(pos + 1);
      String fullTableName = fullSource.substring(0, pos);

      Persistence persistence = findExistingTable(fullTableName, false);
      if (persistence == null)
      {
         return new logical(false);
      }

      String srcTableName = parseTableName(fullTableName)[0];
      P2JField srcField = getExistingField(persistence, srcTableName, srcFieldName);
      if (srcField == null)
      {
         displayCouldNotFindLikeField(srcFieldName);
         return new logical(false);
      }

      P2JField targetField = copyField(targetName, srcField);

      addField(targetField);
      return new logical(true);
   }

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method of Progress 4GL.
    *
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Handle to field name from which to copy the field.
    *
    * @return  <code>true</code> on success.
    */
   public logical addFieldLike(String name, handle source)
   {
      return addFieldLike(new character(name), source);
   }

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Handle to field name from which to copy the field.
    * 
    * @return  <code>true</code> on success.
    */
   public logical addFieldLike(character name, handle source)
   {
      addFunctionCalled = true;

      if (_prepared())
      {
         displayPrepared();
         return new logical(false);
      }

      String targetName = name.getValue();
      if (targetName == null || !source._isValid())
      {
         displayUnableToEvaluate();
         return new logical(false);
      }

      if (hasField(targetName))
      {
         // we have the field with the same name
         return new logical(false);
      }

      Object handleValue = source.get();
      if (!FieldReference.class.isAssignableFrom(handleValue.getClass()))
      {
         ErrorManager.displayError(9056,
                                   "ADD-LIKE-FIELD requires valid BUFFER-FIELD object",
                                   false);
         return new logical(false);
      }

      FieldReference ref = (FieldReference) handleValue;
      RecordBuffer buffer = ref.getParentBuffer();
      String fieldNameHibernate = ref.getPropertyName();

      P2JField srcField = getExistingField(buffer, fieldNameHibernate);
      P2JField targetField = copyField(targetName, srcField);

      addField(targetField);
      return new logical(true);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, String field, String ascending)
   {
      return addFieldToIndex(new character(indexName), 
                             new character(field),
                             new character(ascending));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, String field, character ascending)
   {
      return addFieldToIndex(new character(indexName), new character(field), ascending);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(character indexName, String field, String ascending)
   {
      return addFieldToIndex(indexName, new character(field), new character(ascending));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(character indexName, String field, character ascending)
   {
      return addFieldToIndex(indexName, new character(field), ascending);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, character field, String ascending)
   {
      return addFieldToIndex(new character(indexName), field, ascending);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, character field, character ascending)
   {
      return addFieldToIndex(new character(indexName), field, ascending);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(character indexName, character field, String ascending)
   {
      return addFieldToIndex(indexName, field, new character(ascending));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(character indexName, character field, character ascending)
   {
      addFunctionCalled = true;

      if (_prepared())
      {
         displayPrepared();
         return new logical(false);
      }

      String idxName = indexName.getValue();
      String fieldName = field.getValue();
      if (idxName == null || fieldName == null)
      {
         displayUnableToEvaluate();
         return new logical(false);
      }

      String ascVal = ascending.getValue();
      boolean asc = (ascVal == null);

      if (!asc)
      {
         ascVal = ascVal.toLowerCase();
         asc = ascVal.startsWith("asc");

         if (!asc)
         {
            if (!ascVal.startsWith("desc"))
            {
               ErrorManager.displayError(9062,
                             "Third argument to ADD-INDEX-FIELD must be DESC or ASC, in quotes",
                             false);
               return new logical(false);
            }
         }
      }

      P2JField p2jField = getField(fieldName);

      // field with this name exist...
      boolean invalidField = (p2jField == null);

      P2JIndex index = null;
      if (!invalidField)
      {
         // ...and wasn't already added to the index
         index = getIndex(idxName);
         Iterator<P2JIndexComponent> iter = index.components();
         while (iter.hasNext())
         {
            P2JIndexComponent component = iter.next();
            if (hasSameNormalizedName(component.getName(), fieldName))
            {
               invalidField = true;
               break;
            }
         }
      }

      if (invalidField)
      {
         displayDuplicateOrInvalidIndexField(idxName, fieldName);
         return new logical(false);
      }

      if (p2jField.getExtent() > 1)
      {
         ErrorManager.displayError(370, "Array fields cannot be Index-Fields");
         return new logical(false);
      }

      index.addComponent(fieldName, !asc, false, false, false);
      return new logical(true);
   }

   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, String field)
   {
      return addFieldToIndex(new character(indexName), 
                             new character(field), new character("asc"));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   public logical addFieldToIndex(character indexName, String field)
   {
      return addFieldToIndex(indexName, new character(field), new character("asc"));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   public logical addFieldToIndex(String indexName, character field)
   {
      return addFieldToIndex(new character(indexName), field, new character("asc"));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   public logical addFieldToIndex(character indexName, character field)
   {
      return addFieldToIndex(indexName, field, new character("asc"));
   }

   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the index to which the source index is being copied. Progress ignores
    *          this option and uses the name of the source index instead.
    * @param   sourceName
    *          The name of the index in the source table that is being copied 
    *          to the temp-table.
    * @param   dbTable
    *          The database table name from which to copy the index.
    * 
    * @return  <code>true</code> on success.
    */
   public logical addIndexLike(String name, String sourceName, String dbTable)
   {
      return addIndexLike(new character(name), new character(sourceName), new character(dbTable));
   }
   
   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the index to which the source index is being copied. Progress ignores
    *          this option and uses the name of the source index instead.
    * @param   sourceName
    *          The name of the index in the source table that is being copied 
    *          to the temp-table.
    * @param   dbTable
    *          The database table name from which to copy the index.
    * 
    * @return  <code>true</code> on success.
    */
   public logical addIndexLike(character name,
                               character sourceName,
                               character dbTable)
   {
      addFunctionCalled = true;

      if (_prepared())
      {
         displayPrepared();
         return new logical(false);
      }

      String srcIndexName = sourceName.getValue();
      String fullTableName = dbTable.getValue();
      if (name.isUnknown() || srcIndexName == null || fullTableName == null)
      {
         displayUnableToEvaluate();
         return new logical(false);
      }

      Persistence persistence = findExistingTable(fullTableName, true);
      if (persistence == null)
         return new logical(false);

      String srcTableName = parseTableName(fullTableName)[0];
      P2JIndex srcIndex = getExistingIndex(persistence, srcTableName, srcIndexName);
      if (srcIndex == null)
      {
         displayCouldNotFindLikeIndex(get4GLTableName(persistence, srcTableName), srcIndexName);
         return new logical(false);
      }

      return addIndexLike(srcIndex);
   }

   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to which the source index is being copied. Progress ignores
    *          this option and uses the name of the source index instead.
    * @param   sourceName
    *          The name of the index in the source table that is being copied
    *          to the temp-table.
    * @param   sourceBuffer
    *          Buffer handle from which to copy the index.
    *
    * @return  <code>true</code> on success.
    */
   public logical addIndexLike(String name, String sourceName, handle sourceBuffer)
   {
      return addIndexLike(new character(name), new character(sourceName), sourceBuffer);
   }

   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to which the source index is being copied. Progress ignores
    *          this option and uses the name of the source index instead.
    * @param   sourceName
    *          The name of the index in the source table that is being copied
    *          to the temp-table.
    * @param   sourceBuffer
    *          Buffer handle from which to copy the index.
    *
    * @return  <code>true</code> on success.
    */
   public logical addIndexLike(character name, character sourceName, handle sourceBuffer)
   {
      addFunctionCalled = true;

      if (_prepared())
      {
         displayPrepared();
         return new logical(false);
      }

      String srcIndexName = sourceName.getValue();
      if (name.isUnknown() || srcIndexName == null || !sourceBuffer._isValid())
      {
         displayUnableToEvaluate();
         return new logical(false);
      }

      Object handleValue = sourceBuffer.get();
      if (!Buffer.class.isAssignableFrom(handleValue.getClass()))
      {
         displayCouldNotFindLikeTableFromHandle();
         return new logical(false);
      }

      Buffer buffer = (Buffer) handleValue;
      RecordBuffer recordBuffer = RecordBuffer.get((DataModelObject) buffer);

      P2JIndex srcIndex = getExistingIndex(recordBuffer, srcIndexName);
      if (srcIndex == null)
      {
         Class dmoIface = recordBuffer.getDMOInterface();
         displayCouldNotFindLikeIndex(get4GLTableName(dmoIface), srcIndexName);
         return new logical(false);
      }

      return addIndexLike(srcIndex);
   }

   /**
    * Returns <code>true</code> if this temp-table is in prepared state (after a successful call
    * to {@link #tempTablePrepare}).
    * This method implements the read-only PREPARED attribute from 4GL.
    *
    * @return  <code>true</code> if this table is prepared.
    */
   public logical prepared()
   {
      return new logical(_prepared());
   }

   /**
    * Returns <code>true</code> if this temp-table is in prepared state (after a successful call
    * to {@link #tempTablePrepare}).
    * This method implements the read-only PREPARED attribute from 4GL.
    *
    * @return  <code>true</code> if this table is prepared.
    */
   public boolean _prepared()
   {
      return !super.getName().isUnknown();
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other)
   {
      return copyTempTable(other, new logical(false));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other, logical append)
   {
      return copyTempTable(other, append, new logical(false));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other, logical append, logical replace)
   {
      return copyTempTable(other, append, replace, new logical(false));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other, logical append, logical replace, logical looseCopy)
   {
      return copyTempTable(other, append, replace, looseCopy, new character());
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other, boolean append)
   {
      return copyTempTable(other, new logical(append));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other, boolean append, boolean replace)
   {
      return copyTempTable(other, new logical(append), new logical(replace));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle other, boolean append, boolean replace, boolean looseCopy)
   {
      return copyTempTable(other,
                           new logical(append),
                           new logical(replace),
                           new logical(looseCopy));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    * @param   prefix
    *          Prefix for naming the copied temp-table.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle  other,
                                boolean append,
                                boolean replace,
                                boolean looseCopy,
                                String  prefix)
   {
      return copyTempTable(other,
                           new logical(append),
                           new logical(replace),
                           new logical(looseCopy),
                           new character(prefix));
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    * @param   prefix
    *          Prefix for naming the copied temp-table.
    *
    * @return  true if operation is successful
    */
   public logical copyTempTable(handle    other,
                                logical   append,
                                logical   replace,
                                logical   looseCopy,
                                character prefix)
   {
      // TODO implement actual logic
      return new logical(true);
   }
   /**
    * Sets the UNDO attribute of 4GL temp-table.
    *
    * @param   undo
    *          The new value for the attribute.
    */
   public void setCanUndo(logical undo)
   {
      if (undo.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute UNDO on TEMP-TABLE widget",
               false);
         return;
      }

      this.undoable = undo.booleanValue();

      // TODO UNDO attribute can changed (and will take effect!) after the table has been prepared
   }

   /** 
    * Obtain the current UNDO attribute of 4GL temp-table.
    * 
    * @return  true if the the temp-table can undo.
    */
   public logical canUndo()
   {
      return new logical(undoable);
   }

   /**
    * Sets the ERROR-STRING attribute of 4GL temp-table.
    *
    * @param   errStr
    *          The new value for the ERROR-STRING attribute.
    */
   public void changeErrorString(character errStr)
   {
      if (!_prepared())
      {
         ErrorManager.displayError(12346,
           "ERROR-STRING can only be used with a PREPARED temp-table that is not a BEFORE-TABLE",
           false);
         ErrorManager.recordOrThrowError(3131,
               "Unable to set attribute ERROR-STRING in widget of type TEMP-TABLE",
               false);
         return;
      }

      this.errorString.assign(errStr);
   }

   /** 
    * Obtain the current ERROR-STRING attribute of 4GL temp-table.
    * 
    * @return  the current ERROR-STRING of the temp-table
    */
   public character errorString()
   {
      return errorString;
   }

   /**
    * Checks  whether the resource is dynamic or static.
    * 
    * @return   <code>true</code> if the resource is dynamic, <code>false</code> otherwise. 
    */
   public logical dynamic()
   {
      return new logical(dynamic);
   }

   /**
    * Checks  whether the resource is dynamic or static.
    * 
    * @return   <code>true</code> if the resource is dynamic, <code>false</code> otherwise. 
    */
   public boolean _dynamic()
   {
      return dynamic;
   }
   
   /**
    * Checks whether this temp-table contains any records.
    * 
    * @return  true if this temp-table has records.
    */
   public logical hasRecords()
   {
      if (!_prepared())
         return new logical(false);

      Persistence persistence = PersistenceFactory.getInstance(DatabaseManager.TEMP_TABLE_DB);
      String sql = String.format("select 1 from %s", persistentClass.getTable().getName());
      try
      {
         ScrollableResults res = persistence.executeSQLQuery(sql);
         if (res.first())
            return new logical(true);
      }
      catch (PersistenceException e)
      {
         ErrorManager.recordOrThrowError(e);
         return new logical(false);
      }

      // check for uncommitted records in buffers
      if (handle.unwrapBuffer(defaultBufferHandle)._available())
         return new logical(true);

      // TODO check in other buffers

      return new logical(false);
   }

   /**
    * Returns the temp-table to it's clear state, removing all defined fields, indexes, all
    * records in the table and all corresponding buffers.
    * 
    * @return  true if operation is successful
    */
   public logical clear()
   {
      if (undoable && _prepared() && hasRecords().getValue())
      {
         ErrorManager.displayError(9085,
               "CLEAR method may not be called on a static temp-table object, nor on a dynamic " +
               "one with records; use buffer:EMPTY-TEMP-TABLE",
               false);
         return new logical(false);
      }

      return new logical(forceClear());
   }

   /**
    * Signals that all the field and index definitions for a temp-table 
    * have been provided. After the call to this method no fields and indexes
    * can be added to this temporary table.
    * This method is the P2J equivalent of <code>TEMP-TABLE-PREPARE</code> 
    * method of Progress 4GL.
    * 
    * @param   name
    *          Temp-table name to be used in subsequent query statements that
    *          refer to this temp-table.
    *
    * @return  <code>true</code> on success.
    */
   public logical tempTablePrepare(String name)
   {
      return tempTablePrepare(new character(name));
   }

   /**
    * Signals that all the field and index definitions for a temp-table 
    * have been provided. After the call to this method no fields and indexes
    * can be added to this temporary table.
    * This method is the P2J equivalent of <code>TEMP-TABLE-PREPARE</code> 
    * method of Progress 4GL.
    * 
    * @param   name
    *          Temp-table name to be used in subsequent query statements that
    *          refer to this temp-table.
    *          
    * @return  <code>true</code> on success.
    */
   public logical tempTablePrepare(character name)
   {
      if (_prepared())
      {
         ErrorManager.displayError(9043, "TEMP-TABLE is already prepared--ignoring", false);
         return new logical(false);
      }

      String tableName = name.getValue();
      if (tableName == null)
      {
         ErrorManager.displayError(9046, "Unable to PREPARE TEMP-TABLE buffer object", false);
         forceClear();
         return new logical(false);
      }

      if (fields.size() == 0)
      {
         ErrorManager.displayError(9033,
                                   "TEMP-TABLE object must have at least one field before " +
                                      "TEMP-TABLE-PREPARE can be called",
                                   false);
         displayUnableToPrepareTableFields();
         forceClear();
         return new logical(false);
      }

      int nameLength = tableName.length();
      if (nameLength == 0 || nameLength > 32)
      {
         ErrorManager.displayError(1700,
                                   "Identifier was left blank or is more than 32 characters",
                                   false);
         displayUnableToPrepareTableFields();
         forceClear();
         return new logical(false);
      }

      if (!Character.isLetter(tableName.charAt(0)))
      {
         ErrorManager.displayError(257,
                     String.format("The first character of %s must be alphabetic", tableName));
         displayUnableToPrepareTableFields();
         forceClear();
         return new logical(false);
      }

      for (int i = 1; i < nameLength; i++)
      {
         char c = tableName.charAt(i);
         if (!Character.isLetter(c) && !Character.isDigit(c) && !allowedSpecialChars.contains(c))
         {
            ErrorManager.displayError(274,
                     String.format("The character %s is not permitted in name %s", c, tableName));
            displayUnableToPrepareTableFields();
            forceClear();
            return new logical(false);
         }
      }

      boolean foundPrimary = false;
      for (P2JIndex index : indexes.values())
      {
         if (index.isPrimary())
         {
            if (foundPrimary)
            {
               // there are multiple primary indexes
               ErrorManager.displayError(3389,
                                        "TEMP-TABLE  must have only one primary key",
                                        false);
               displayUnableToPrepareTableIndices();
               forceClear();
               return new logical(false);
            }
            else
            {
               foundPrimary = true;
            }
         }

         if (index.size() == 0)
         {
            ErrorManager.displayError(9029,
               String.format("TEMP-TABLE index %s needs at least 1 component", index.getName()),
               false);
            displayUnableToPrepareTableIndices();
            forceClear();
            return new logical(false);
         }
      }

      boolean res = DynamicTablesHelper.getInstance().createDynamicDMO(this);
      if (res)
      {
         createDefaultBuffer();
         super.setName(name); // also makes table "prepared"
      }
      else
      {
         forceClear();
      }

      return new logical(res);
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL.
    *
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(String tableName)
   {
      return addFieldsFrom(new character(tableName));
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL.
    *
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the field definitions.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(handle hbuf)
   {
      return addFieldsFrom(hbuf, (String) null);
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL.
    *
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(character tableName)
   {
      return addFieldsFrom(tableName, new character());
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL.
    *
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the field definitions.
    * @param   except
    *          A comma-separated list of field names to be excluded from the copy process.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(handle hbuf, String except)
   {
      return addFieldsFrom(hbuf, new character(except));
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL.
    *
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    * @param   except
    *          A comma-separated list of field names to be excluded from the copy process.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(String tableName, String except)
   {
      return addFieldsFrom(new character(tableName), new character(except));
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL.
    *
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    * @param   except
    *          A comma-separated list of field names to be excluded from the copy process.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(String tableName, character except)
   {
      return addFieldsFrom(new character(tableName), except);
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL.
    *
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the field definitions.
    * @param   except
    *          A comma-separated list of field names to be excluded from the copy process.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(handle hbuf, character except)
   {
      if (_prepared())
      {
         displayPrepared();
         return new logical(); // unknown value is returned
      }

      RecordBuffer recordBuffer = getRecordBufferByHandle(hbuf, true);
      if (recordBuffer == null)
         return new logical(false);

      doAddFieldsFrom(getExistingFields(recordBuffer), except);
      return new logical(true);
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL.
    *
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    * @param   except
    *          A comma-separated list of field names to be excluded from the copy process.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(character tableName, character except)
   {
      addFunctionCalled = true;

      if (_prepared())
      {
         displayPrepared();
         return new logical(); // unknown value is returned
      }

      String fullTableName = tableName.getValue();
      if (fullTableName == null)
      {
         displayUnableAddFieldsFrom();
         return new logical(false);
      }

      Persistence persistence = findExistingTable(fullTableName, true);
      if (persistence == null)
      {
         displayUnableAddFieldsFrom();
         return new logical(false);
      }

      String tblName = parseTableName(fullTableName)[0];
      doAddFieldsFrom(getExistingFields(persistence, tblName), except);
      return new logical(true);
   }

   /**
    * Obtain the handle to the buffer that was created by default for this 
    * temporary table. Every dynamic temp-table is created with at 
    * least one buffer. This buffer's object handle is returned by this
    * method. It cannot be called until the table has been prepared using 
    * <code>tempTablePrepare()</code> method.
    * This method is the P2J equivalent of <code>DEFAULT-BUFFER-HANDLE</code>
    * attribute.
    * 
    * @return  The default buffer handle for the temporary table.
    */
   public handle defaultBufferHandle()
   {
      if (!_prepared())
      {
         displayNotPrepared();
         return new handle();
      }

      return defaultBufferHandle;
   }

   /**
    * Set table NAME attribute. Note that it can be changed only after the table has been
    * prepared.
    *
    * @param name
    *        New table name.
    */
   public void setName(character name)
   {
      if (name.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute NAME on TEMP-TABLE widget",
               false);
         return;
      }

      if (!_prepared())
      {
         displayNotPrepared();
         ErrorManager.recordOrThrowError(3131,
               "Unable to set attribute NAME in widget of type TEMP-TABLE",
               false);
         return;
      }

      super.setName(name);
   }

   /**
    * Get table NAME attribute.
    *
    * @return  name
    *          Table NAME attribute value. If the table hasn't been prepared, unknown value is
    *          returned.
    */
   public character getName()
   {
      character name = super.getName();
      if (name.isUnknown())
      {
         displayNotPrepared();
      }

      return name;
   }

   /**
    * Returns the state of the ERROR attribute
    *
    * @return    Current state of the ERROR attribute.
    */
   public logical isError()
   {
      if (!_prepared())
      {
         displayNotPrepared();
      }

      return error;
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(boolean value)
   {
      setError(new logical(value));
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute ERROR on TEMP-TABLE widget",
               false);
         return;
      }

      if (!_prepared())
      {
         displayNotPrepared();
         ErrorManager.recordOrThrowError(3131,
               "Unable to set attribute ERROR in widget of type TEMP-TABLE",
               false);
         return;
      }

      error.assign(value);
   }

   /**
    * Check if this resource supports the PREV-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasPrevSibling()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the NEXT-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasNextSibling()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the NAME attribute only in read-only mode.
    * 
    * @return   Always <code>false</code>, as NAME attribute is writable.
    */
   @Override
   protected boolean hasNameReadOnly()
   {
      return false;
   }

   /**
    * Gets the the unique ID number associated to this object by the
    * underlying system. Not the same as the handle value itself.
    * 
    * @return   The integer number of the children.
    */
   public integer getUniqueID()
   {
      return uniqueId;
   }

   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return true;
   }

   /**
    * Worker to be implemented by each resource.  Called by {@link #delete()}.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   protected boolean resourceDelete()
   {
      if (!implicitDeletion() && !_dynamic())
      {
         // static temp-tables can not be deleted, but no error message is shown
         return false;
      }
      
      return forceClear();
   }
   
   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   public character getADMData()
   {
      return admData;
   }

    /**
     * Set the value of the ADM-DATA attribute.
     * 
     * @param    value
     *           The new value.
     */
   public void setADMData(String value)
   {
      setADMData(new character(value));
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   public void setADMData(character value)
   {
      admData.assign(value);
   }

   /**
    * Get the collection of table fields.
    *
    * @return  the collection of table fields.
    */
   public Collection<P2JField> getFields()
   {
      return fields.values();
   }

   /**
    * Get field by its name.
    *
    * @param  fieldName
    *         Field name.
    *
    * @return field with the specified name (normalized field names are compared).
    */
   public P2JField getField(String fieldName)
   {
      return fields.get(normalizeName(fieldName));
   }

   /**
    * Get index by its name.
    *
    * @param  indexName
    *         Index name.
    *
    * @return index with the specified name (normalized index names are compared).
    */
   public P2JIndex getIndex(String indexName)
   {
      return indexes.get(normalizeName(indexName));
   }

   /**
    * Get table indexes.
    *
    * @return see above.
    */
   public Collection<P2JIndex> getIndexes()
   {
      return indexes.values();
   }

   /**
    * Set DMO interface class which corresponds the table.
    *
    * @param dmoIface
    *        DMO interface class which corresponds the table.
    */
   void setDmoIface(Class dmoIface)
   {
      this.dmoIface = dmoIface;
   }

   /**
    * Set persistent class which corresponds the table.
    *
    * @param persistentClass
    *        Persistent class which corresponds the table.
    */
   void setPersistentClass(PersistentClass persistentClass)
   {
      String tableName = persistentClass.getTable().getName();
      for (P2JIndex index : indexes.values())
      {
         index.setTable(tableName);
      }

      this.persistentClass = persistentClass;
   }

   /**
    * Get persistent class which corresponds the table.
    *
    * @return persistent class which corresponds the table or <code>null</code> if the table is
    *         not prepared.
    */
   PersistentClass getPersistentClass()
   {
      return persistentClass;
   }

   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   tableName
    *          The name of a table to create the copy of.
    * @param   indexName
    *          The name of a single index which is copied from the source table.
    * @param   handleIndex
    *          <code>true</code> if <code>indexName</code> parameter should be taken into
    *          consideration. This parameter is required because <code>unknown</code> index name
    *          represents an index with empty name rather than "index not specified" case.
    *
    * @return  <code>true</code> on success.
    */
   private logical createTableLikeImpl(character tableName,
                                       character indexName,
                                       boolean handleIndex)
   {
      if (addFunctionCalled)
      {
         displayCreateLikeNotFirst();
         return new logical();  // unknown is returned
      }

      addFunctionCalled = true;

      String fullTableName = tableName.getValue();
      if (fullTableName == null)
      {
         displayUnableCreateLike();
         return new logical(false);
      }

      Persistence persistence = findExistingTable(fullTableName, true);
      if (persistence == null)
      {
         displayUnableCreateLike();
         return new logical(false);
      }

      String tblName = parseTableName(fullTableName)[0];
      addAllFields(getExistingFields(persistence, tblName));

      if (handleIndex)
      {
         String idxName = indexName.isUnknown() ? "" : indexName.getValue(); // correct
         P2JIndex srcIndex = getExistingIndex(persistence, tblName, idxName);
         if (srcIndex == null)
         {
            displayUnableToFindIndex(get4GLTableName(persistence, tblName), idxName);
            displayUnableCreateLike();
            return new logical(false);
         }
         else
         {
            addIndex(srcIndex);
            return new logical(true);
         }
      }
      else
      {
         addAllIndexes(getExistingIndexes(persistence, tblName));
         return new logical(true);
      }
   }

   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the definitions.
    * @param   indexName
    *          The name of a single index which is copied from the source table.
    * @param   handleIndex
    *          <code>true</code> if <code>indexName</code> parameter should be taken into
    *          consideration. This parameter is required because <code>unknown</code> index name
    *          represents an index with empty name rather than "index not specified" case.
    *
    * @return  <code>true</code> on success.
    */
   private logical createTableLikeImpl(handle hbuf, character indexName, boolean handleIndex)
   {
      if (addFunctionCalled)
      {
         displayCreateLikeNotFirst();
         return new logical();  // unknown is returned
      }

      addFunctionCalled = true;

      RecordBuffer recordBuffer = getRecordBufferByHandle(hbuf, false);
      if (recordBuffer == null)
         return new logical(false);

      addAllFields(getExistingFields(recordBuffer));

      if (handleIndex)
      {
         String idxName = indexName.isUnknown() ? "" : indexName.getValue(); // correct
         P2JIndex srcIndex = getExistingIndex(recordBuffer, idxName);
         if (srcIndex == null)
         {
            displayUnableToFindIndex(get4GLTableName(recordBuffer.getDMOInterface()), idxName);
            displayUnableCreateLike();
            return new logical(false);
         }
         else
         {
            addIndex(srcIndex);
            return new logical(true);
         }
      }
      else
      {
         addAllIndexes(getExistingIndexes(recordBuffer));
         return new logical(true);
      }
   }

   /**
    * Adds to the temp-table an index like the one specified. This method is the P2J equivalent
    * of <code>ADD-LIKE-FIELD</code> method of Progress 4GL.
    *
    * @param   srcIndex
    *          The index in the source table that is being copied to the temp-table.
    *
    * @return  <code>true</code> on success.
    */
   private logical addIndexLike(P2JIndex srcIndex)
   {
      String srcIndexName = srcIndex.getName();

      if (hasIndex(srcIndexName))
      {
         displayDuplicateIndex(srcIndexName);
         displayUnableToCreateLikeIndex();
         displayUnableToCreateLikeIndex(); // this message is displayed twice
         return new logical(false);
      }

      P2JIndex targetIndex = copyIndex(srcIndex);
      boolean res = true;

      Iterator<P2JIndexComponent> components = srcIndex.components();
      while (components.hasNext())
      {
         P2JIndexComponent srcComponent = components.next();
         String componentName = srcComponent.getName();
         if (!hasField(componentName))
         {
            displayDuplicateOrInvalidIndexField(srcIndexName, componentName);
            displayUnableToCreateLikeIndex();
            displayUnableToCreateLikeIndex(); // this message is displayed twice
            res = false;
            break;
         }
         else
         {
            targetIndex.addComponent(srcComponent.getName(),
                                     srcComponent.isDescending(),
                                     srcComponent.isCharType(),
                                     srcComponent.isIgnoreCase(),
                                     srcComponent.isAbbreviated());
         }
      }

      addIndex(targetIndex);
      return new logical(res);
   }

   /**
    * Add specified set of fields to the table. If a field is in the except-list or was already
    * added to the table, it is ignored.
    * 
    * @param srcFields
    *        Set of fields to be added.
    * @param except
    *        Comma-separated list of field names that should be excluded from the copy process.
    */
   private void doAddFieldsFrom(Collection<P2JField> srcFields, character except)
   {
      if (srcFields != null)
      {
         Set<String> exceptList = parseFieldList(except);

         for (P2JField srcField : srcFields)
         {
            // names in except list should match original field names, case sensitivity isn't
            // taken into consideration, but spaces are taken into consideration
            String fieldName = srcField.getName().toLowerCase();
            if (!hasField(fieldName) && (exceptList == null || !exceptList.contains(fieldName)))
               addField(srcField);
         }
      }
   }

   /**
    * Create default buffer for the table.
    */
   private void createDefaultBuffer()
   {
      if (defaultBufferHandle == null)
      {
         Temporary buffer =
               DynamicTablesHelper.getInstance().createDynamicBuffer(dmoIface,
                                                            persistentClass.getTable().getName(),
                                                            undoable);
         defaultBufferHandle = new handle();
         defaultBufferHandle.assign(buffer);
      }
   }

   /**
    * Validates data type name which was passed to ADD-NEW-FIELD function.
    *
    * @param  datatype
    *         Data type name.
    *
    * @return corresponding {@link ParmType} or <code>null</code> if validation has failed.
    */
   private ParmType validateDatatype(String datatype)
   {
      // get first word
      int spaceIdx = datatype.indexOf(' ');
      String compareStr = spaceIdx >= 0 ? datatype.substring(0, spaceIdx) : datatype;

      if (compareStr.length() > 0)
      {
         compareStr = compareStr.toLowerCase();

         for (String type : datatypesMatchLen.keySet())
         {
            if (type.startsWith(compareStr))
            {
               if (compareStr.length() >= datatypesMatchLen.get(type))
               {
                  ParmType res = ParmType.fromString(type);
                  if (res == null && DATETIMETZ_TYPE_NAME.equals(type))
                  {
                     res = ParmType.DTTZ;
                  }
                  return res;
               }
            }
         }
      }

      ErrorManager.displayError(9051,
                            String.format("Illegal datatype given in ADD-NEW-FIELD %s", datatype),
                            false);
      return null;
   }

   /**
    * Display "Unable to prepare TEMP-TABLE fields" error message.
    */
   private void displayUnableToPrepareTableFields()
   {
      ErrorManager.displayError(9044, "Unable to prepare TEMP-TABLE fields", false);
   }

   /**
    * Display "Unable to prepare TEMP-TABLE indices" error message.
    */
   private void displayUnableToPrepareTableIndices()
   {
      ErrorManager.displayError(9045, "Unable to prepare TEMP-TABLE indices", false);
   }

   /**
    * Display "TEMP-TABLE must be PREPARED before referencing NAME/DEFAULT-BUFFER-HANDLE etc"
    * error message.
    */
   private void displayNotPrepared()
   {
      ErrorManager.displayError(9040,
               "TEMP-TABLE must be PREPARED before referencing NAME/DEFAULT-BUFFER-HANDLE etc",
               false);

   }

   /**
    * Display "Definitional TEMP-TABLE methods (ADD-NEW/LIKE) may not be called after
    * TEMP-TABLE-PREPARE has been called unless a CLEAR is done first, and may never be called on
    * a DEFINED static table" error message.
    */
   private void displayPrepared()
   {
      ErrorManager.displayError(9042,
               "Definitional TEMP-TABLE methods (ADD-NEW/LIKE) may not be called after " +
               "TEMP-TABLE-PREPARE has been called unless a CLEAR is done first, and may never " +
               "be called on a DEFINED static table",
               false);

   }

   /**
    * Display "Unable to evaluate arguments for TEMP-TABLE method" error message.
    */
   private void displayUnableToEvaluate()
   {
      ErrorManager.displayError(9050,
                                "Unable to evaluate arguments for TEMP-TABLE method",
                                false);
   }

   /**
    * Display "Could not find table {tableName} during ADD-LIKE-FIELD. Table names must not be
    * abbreviated." error message.
    *
    * @param tableName
    *        Table name.
    */
   private void displayCouldNotFindLikeTable(String tableName)
   {
      ErrorManager.displayError(9054,
                                String.format("Could not find table %s during ADD-LIKE-FIELD." +
                                             "  Table names must not be abbreviated", tableName),
                                false);
   }

   /**
    * Display "TEMP-TABLE method could not find LIKE table from BUFFER handle" error message.
    */
   private void displayCouldNotFindLikeTableFromHandle()
   {
      ErrorManager.displayError(9058,
                                "TEMP-TABLE method could not find LIKE table from BUFFER handle",
                                false);
   }

   /**
    * Display "TEMP-TABLE method could not find LIKE table name {tableName}}" error message.
    *
    * @param tableName
    *        Table name.
    */
   private void displayCouldNotFindLikeTableName(String tableName)
   {
      ErrorManager.displayError(9057,
                 String.format("TEMP-TABLE method could not find LIKE table name %s", tableName),
                 false);
   }

   /**
    * Display "Duplicate index name {indexName} for temp-table" error message.
    *
    * @param indexName
    *        Index name.
    */
   private void displayDuplicateIndex(String indexName)
   {
      ErrorManager.displayError(3390,
                             String.format("Duplicate index name %s for temp-table ", indexName),
                             false);
   }

   /**
    * Display "Could not find source field {fieldName} during ADD-LIKE-FIELD." error message.
    *
    * @param fieldName
    *        Field name.
    */
   private void displayCouldNotFindLikeField(String fieldName)
   {
      ErrorManager.displayError(9055,
               String.format("Could not find source field %s during ADD-LIKE-FIELD", fieldName),
               false);
   }

   /**
    * Display "ADD-LIKE-INDEX could not find index {indexName} in table {tableName}" error
    * message.
    *
    * @param tableName
    *        Table name.
    * @param indexName
    *        Index name.
    */
   private void displayCouldNotFindLikeIndex(String tableName, String indexName)
   {
      ErrorManager.displayError(9059,
                              String.format("ADD-LIKE-INDEX could not find index %s in table %s",
                                            indexName,
                                            tableName),
                              false);
   }

   /**
    * Display "Unable to create index for TEMP-TABLE in ADD-LIKE-INDEX" error message.
    */
   private void displayUnableToCreateLikeIndex()
   {
      ErrorManager.displayError(9060,
                                "Unable to create index for TEMP-TABLE in ADD-LIKE-INDEX",
                                false);
   }

   /**
    * Display "Attempt to add duplicate or inappropriate word index field {fieldName} to index
    * {indexName}" error message.
    *
    * @param indexName
    *        Index name.
    * @param fieldName
    *        Field name.
    */
   private void displayDuplicateOrInvalidIndexField(String indexName, String fieldName)
   {
      ErrorManager.displayError(9063,
       String.format("Attempt to add duplicate or inappropriate word index field %s to index %s",
                     fieldName,
                     indexName),
                     false);
   }

   /**
    * Display "Unable to CREATE-LIKE for TEMP-TABLE--try using a smaller source table" error
    * message.
    */
   private void displayUnableCreateLike()
   {
      ErrorManager.displayError(9047,
                    "Unable to CREATE-LIKE for TEMP-TABLE--try using a smaller source table",
                    false);
   }

   /**
    * Display "CREATE-LIKE must be the first method on a newly CREATEd TEMP-TABLE, or a newly
    * CLEARed one" error message.
    */
   private void displayCreateLikeNotFirst()
   {
      ErrorManager.displayError(9041,
                    "CREATE-LIKE must be the first method on a newly CREATEd TEMP-TABLE, " +
                    "or a newly CLEARed one",
                    false);
   }

   /**
    * Display "Unable to ADD-FIELDS-FROM for TEMP-TABLE" error message.
    */
   private void displayUnableAddFieldsFrom()
   {
      ErrorManager.displayError(9049, "Unable to ADD-FIELDS-FROM for TEMP-TABLE", false);
   }

   /**
    * Display "Unable to find index {indexName} in table {tableName}" error message.
    *
    * @param tableName
    *        Table name.
    * @param indexName
    *        Index name.
    */
   private void displayUnableToFindIndex(String tableName, String indexName)
   {
      ErrorManager.displayError(9048,
                  String.format("Unable to find index %s in table %s", indexName, tableName),
                  false);
   }

   /**
    * Unconditionally delete all buffers for this table, drop DMO definitions and temporary table,
    * clear TEMP-TABLE data structures.
    *
    * @return <code>true</code> if operation was successful.
    */
   private boolean forceClear()
   {
      if (_prepared())
      {
         DynamicTablesHelper helper = DynamicTablesHelper.getInstance();

         // delete all records
         if (defaultBufferHandle != null)
         {
            Buffer buffer = handle.unwrapBuffer(defaultBufferHandle);
            buffer.release();
            ((TemporaryBuffer) RecordBuffer.get((DataModelObject) buffer)).dropTable();
            ((Deletable) buffer).delete();

            defaultBufferHandle = null;

            // TODO delete all other buffers for this table
         }

         helper.dropDynamicDMO(dmoIface, true);
      }

      // keep UNIQUE-ID attribute
      addFunctionCalled = false;
      undoable = true; // however default value is "false"
      fields.clear();
      indexes.clear();
      errorString().setUnknown();
      error.setUnknown();
      admData.setUnknown();
      persistentClass = null;
      dmoIface = null;
      super.setName(new character());

      return true;
   }

   /**
    * Get the field of an existing table.
    *
    * @param  persistence
    *         Persistence which corresponds the target database.
    * @param  tableName4GL
    *         4GL table name.
    * @param  fieldName4GL
    *         4GL field name.
    *
    * @return field with the specified name or <code>null</code> if there is no such field.
    */
   private P2JField getExistingField(Persistence persistence,
                                     String tableName4GL,
                                     String fieldName4GL)
   {
      tableName4GL = normalizeName(tableName4GL);
      fieldName4GL = normalizeName(fieldName4GL);

      // TODO implement using MetaSchema. It is assumed that for temp tables in the case of name
      // conflict (there are multiple tables with the same name in different procedures) the table
      // which is reachable from the current scope will be returned. The following code returns
      // data for sample tables and allows dt-* testcases to pass.
      if ((persistence.getDatabase().getName().equals("p2j_test") &&
           tableName4GL.equals("book") &&
           fieldName4GL.equals("book-id")) ||

             (persistence.getDatabase().equals(DatabaseManager.TEMP_TABLE_DB) &&
              tableName4GL.equals("dummy") &&
              fieldName4GL.equals("dummy-field")) ||

            (persistence.getDatabase().equals(DatabaseManager.TEMP_TABLE_DB) &&
              tableName4GL.equals("tt") &&
              (fieldName4GL.equals("f1") || fieldName4GL.equals("f2"))))
         return new P2JField(fieldName4GL, ParmType.INT, 1, "", new integer(), "", "");

      return null;
   }

   /**
    * Get the field of an existing table.
    *
    * @param  buffer
    *         A record buffer which backs records of the target table.
    * @param  fieldNameHibernate
    *         Hibernate field name.
    *
    * @return field with the specified name or <code>null</code> if there is no such field.
    */
   private P2JField getExistingField(RecordBuffer buffer,
                                     String fieldNameHibernate)
   {
      if (buffer.isDynamic())
      {
         Class dmoIface = buffer.getDMOInterface();
         TempTableBuilder table = DynamicTablesHelper.getTableInfo(dmoIface);
         String fieldName4GL = DynamicTablesHelper.get4GLFieldName(dmoIface, fieldNameHibernate);
         return table.getField(fieldName4GL);
      }
      else
      {
         // TODO implement using MetaSchema. It is assumed that for temp tables in the case of
         // name conflict (there are multiple tables with the same name in different procedures)
         // the table which is reachable from the current scope will be returned. The following
         // code returns data for sample tables and allows dt-* testcases to pass.
         return new P2JField(fieldNameHibernate, ParmType.INT, 1, "", new integer(), "", "");
      }
   }

   /**
    * Get the index of an existing table.
    *
    * @param  buffer
    *         A record buffer which backs records of the target table.
    * @param  indexName4GL
    *         4GL index name.
    *
    * @return index with the specified name or <code>null</code> if there is no such index.
    */
   private P2JIndex getExistingIndex(RecordBuffer buffer,
                                     String indexName4GL)
   {
      if (buffer.isDynamic())
      {
         Class dmoIface = buffer.getDMOInterface();
         TempTableBuilder table = DynamicTablesHelper.getTableInfo(dmoIface);
         return table.getIndex(indexName4GL);
      }
      else
      {
         // TODO implement using MetaSchema. It is assumed that for temp tables in the case of
         // name conflict (there are multiple tables with the same name in different procedures)
         // the table which is reachable from the current scope will be returned. The following
         // code returns data for sample tables and allows dt-* testcases to pass.
         P2JIndex index = new P2JIndex(null, indexName4GL, false);
         if (indexName4GL.equals("index1"))
         {
            index.addComponent("f1");
            index.addComponent("f2");
            return index;
         }
         else if (indexName4GL.equals("index2"))
         {
            index.addComponent("f3");
            index.addComponent("f4");
            return index;
         }
         else if (indexName4GL.equals("index3"))
         {
            index.addComponent("f5");
            return index;
         }

         return null;
      }
   }

   /**
    * Get the indexes of an existing table.
    *
    * @param  persistence
    *         Persistence which corresponds the target database.
    * @param  tableName4GL
    *         4GL table name.
    *
    * @return collection of the indexes of the target table or <code>null</code> if there is no
    *         such table.
    */
   private Collection<P2JIndex> getExistingIndexes(Persistence persistence,
                                                   String tableName4GL)
   {
      // TODO implement using MetaSchema. It is assumed that for temp tables in the case of
      // name conflict (there are multiple tables with the same name in different procedures)
      // the table which is reachable from the current scope will be returned. The following
      // code returns data for sample tables and allows dt-* testcases to pass.
      if (tableName4GL.equals("tt"))
      {
         List<P2JIndex> res = new ArrayList<P2JIndex>();

         P2JIndex index = new P2JIndex(null, "index1", false);
         index.addComponent("f1");
         index.addComponent("f2");
         res.add(index);

         index = new P2JIndex(null, "index2", false);
         index.addComponent("f3");
         index.addComponent("f4");
         res.add(index);

         index = new P2JIndex(null, "index3", false);
         index.addComponent("f5");
         res.add(index);

         return res;
      }

      return null;
   }

   /**
    * Get the indexes of an existing table.
    *
    * @param  buffer
    *         A record buffer which backs records of the target table.
    *
    * @return collection of the indexes of the target table or <code>null</code> if there is no
    *         such table.
    */
   private Collection<P2JIndex> getExistingIndexes(RecordBuffer buffer)
   {
      if (buffer.isDynamic())
      {
         Class dmoIface = buffer.getDMOInterface();
         TempTableBuilder table = DynamicTablesHelper.getTableInfo(dmoIface);
         return table.getIndexes();
      }
      else
      {
         // TODO implement using MetaSchema. It is assumed that for temp tables in the case of
         // name conflict (there are multiple tables with the same name in different procedures)
         // the table which is reachable from the current scope will be returned. The following
         // code returns data for sample tables and allows dt-* testcases to pass.
         return getExistingIndexes(null, buffer.getTable());
      }
   }

   /**
    * Get the fields of an existing table.
    *
    * @param  buffer
    *         A record buffer which backs records of the target table.
    *
    * @return collection of the fields of the target table or <code>null</code> if there is no
    *         such table.
    */
   private Collection<P2JField> getExistingFields(RecordBuffer buffer)
   {
      if (buffer.isDynamic())
      {
         Class dmoIface = buffer.getDMOInterface();
         TempTableBuilder table = DynamicTablesHelper.getTableInfo(dmoIface);
         return table.getFields();
      }
      else
      {
         // TODO implement using MetaSchema. It is assumed that for temp tables in the case of
         // name conflict (there are multiple tables with the same name in different procedures)
         // the table which is reachable from the current scope will be returned. The following
         // code returns data for sample tables and allows dt-* testcases to pass.
         return getExistingFields(null, buffer.getTable());
      }
   }

   /**
    * Get the fields of an existing table.
    *
    * @param  persistence
    *         Persistence which corresponds the target database.
    * @param  tableName4GL
    *         4GL table name.
    *
    * @return collection of the fields of the target table or <code>null</code> if there is no
    *         such table.
    */
   private Collection<P2JField> getExistingFields(Persistence persistence,
                                                  String tableName4GL)
   {
      // TODO implement using MetaSchema. It is assumed that for temp tables in the case of
      // name conflict (there are multiple tables with the same name in different procedures)
      // the table which is reachable from the current scope will be returned. The following
      // code returns data for sample tables and allows dt-* testcases to pass.
      if (tableName4GL.equals("tt"))
      {
         List<P2JField> res = new ArrayList<P2JField>();
         res.add(new P2JField("f1", ParmType.INT, 1, "", new integer(), "", ""));
         res.add(new P2JField("f2", ParmType.INT, 1, "", new integer(), "", ""));
         res.add(new P2JField("f3", ParmType.INT, 1, "", new integer(), "", ""));
         res.add(new P2JField("f4", ParmType.INT, 1, "", new integer(), "", ""));
         res.add(new P2JField("f5", ParmType.INT, 1, "", new integer(), "", ""));
         return res;
      }

      return null;
   }

   /**
    * Get the index of an existing table.
    *
    * @param  persistence
    *         Persistence which corresponds the target database.
    * @param  tableName4GL
    *         4GL table name.
    * @param  indexName4GL
    *         4GL index name.
    *
    * @return index with the specified name or <code>null</code> if there is no such index.
    */
   private P2JIndex getExistingIndex(Persistence persistence,
                                     String tableName4GL,
                                     String indexName4GL)
   {
      // TODO implement using MetaSchema. It is assumed that for temp tables in the case of
      // name conflict (there are multiple tables with the same name in different procedures)
      // the table which is reachable from the current scope will be returned. The following
      // code returns data for sample tables and allows dt-* testcases to pass.
      if ((persistence.getDatabase().getName().equals("p2j_test") &&
           tableName4GL.equals("book") &&
           indexName4GL.startsWith("book-index")) ||

             (persistence.getDatabase().equals(DatabaseManager.TEMP_TABLE_DB) &&
              tableName4GL.equals("dummy") &&
              indexName4GL.equals("dummyindex")))
         return new P2JIndex(null, indexName4GL, false);

      if (persistence.getDatabase().equals(DatabaseManager.TEMP_TABLE_DB) &&
          tableName4GL.equals("tt"))
      {
         P2JIndex index = new P2JIndex(null, indexName4GL, false);

         if (indexName4GL.equals("index1"))
         {
            index.addComponent("f1");
            index.addComponent("f2");
            return index;
         }

         if (indexName4GL.equals("index2"))
         {
            index.addComponent("f3");
            index.addComponent("f4");
            return index;
         }

         if (indexName4GL.equals("index3"))
         {
            index.addComponent("f5");
            return index;
         }

         return null;
      }

      if (persistence.getDatabase().equals(DatabaseManager.TEMP_TABLE_DB) &&
          tableName4GL.equals("tt2"))
      {
         P2JIndex index = new P2JIndex(null, indexName4GL, false);

         if (indexName4GL.equals("index1"))
         {
            index.addComponent("f1");
            index.addComponent("f2");
            index.addComponent("fno");
            index.addComponent("f3");
            return index;
         }

         if (indexName4GL.equals("index2"))
         {
            index.addComponent("fno");
            index.addComponent("f3");
            return index;
         }

         return null;
      }

      if (persistence.getDatabase().equals(DatabaseManager.TEMP_TABLE_DB) &&
          tableName4GL.equals("tt3"))
      {
         P2JIndex index = new P2JIndex(null, indexName4GL, false);

         if (indexName4GL.equals("index1"))
         {
            index.addComponent("f1");
            index.addComponent("f2");
            return index;
         }

         if (indexName4GL.equals("index4"))
         {
            index.addComponent("f3");
            index.addComponent("f4");
            return index;
         }

         if (indexName4GL.equals("index3"))
         {
            index.addComponent("f5");
            return index;
         }

         return null;
      }

      return null;
   }

   /**
    * Get record buffer by buffer or temp-table handle. Displays error if the handle is not valid
    * or of a wrong type.
    *
    * @param  hbuf
    *         Buffer or temp-table handle.
    * @param  addFieldsFrom
    *         <code>true</code> if this function is called from ADD-FIELDS-FROM,
    *         <code>false</code> if it is called from CREATE-LIKE. Affects error messaging.
    *
    * @return target record buffer or <code>null</code> if the case of error.
    */
   private RecordBuffer getRecordBufferByHandle(handle hbuf, boolean addFieldsFrom)
   {
      if (!hbuf._isValid())
      {
         if (addFieldsFrom)
            displayUnableAddFieldsFrom();
         else
            displayUnableCreateLike();
         return null;
      }

      Object handleValue = hbuf.get();
      Class handleClass = handleValue.getClass();
      Buffer buffer;

      if (Buffer.class.isAssignableFrom(handleClass))
      {
          buffer = (Buffer) handleValue;
      }
      else if (TempTableBuilder.class.isAssignableFrom(handleClass))
      {
         TempTableBuilder builder = (TempTableBuilder) handleValue;
         if (builder.defaultBufferHandle == null)
         {
            displayCouldNotFindLikeTableFromHandle();
            if (addFieldsFrom)
               displayUnableAddFieldsFrom();
            else
               displayUnableCreateLike();
            return null;
         }

         buffer = handle.unwrapBuffer(builder.defaultBufferHandle);
      }
      else
      {
         displayCouldNotFindLikeTableFromHandle();
         if (addFieldsFrom)
            displayUnableAddFieldsFrom();
         else
            displayUnableCreateLike();
         return null;
      }

      return RecordBuffer.get((DataModelObject) buffer);
   }

   /**
    * Split comma-separated list of field names into the set of separate field names.
    *
    * @param  list
    *         Comma-separated list of fields.
    *
    * @return The set of separate field names or <code>null</code> if there are no fields in the
    *         input list. Names are returned in lower case, spaces are kept.
    */
   private Set<String> parseFieldList(character list)
   {
      if (list.isUnknown())
         return null;

      String val = list.getValue();
      if (val.length() == 0)
         return null;

      String[] splited = val.toLowerCase().split(",");
      Set<String> res = new HashSet<String>();
      res.addAll(Arrays.asList(splited));
      return res;
   }

   /**
    * Search for an existing table with the specified name. Displays error message in the case of
    * error.
    *
    * @param  fullTableName
    *         Table name in format "databaseName.tableName" or "tableName". In the latter case
    *         table search is performed in all permanent databases and then the temporary
    *         database. Dynamic tables are ignored.
    * @param  tableNameSearch
    *         <code>true</code> if it is called from ADD-LIKE-INDEX, ADD-FIELDS-FROM or
    *         CREATE-LIKE, <code>false</code> if it is called from ADD-FIELD-LIKE. Affects
    *         error messaging.
    *
    * @return persistence which corresponds the target database or <code>null</code> if there is
    *         no such table.
    */
   private Persistence findExistingTable(String fullTableName, boolean tableNameSearch)
   {
      Persistence persistence = DynamicTablesHelper.findExistingTable(fullTableName);
      if (persistence == null)
      {
         if (tableNameSearch)
            displayCouldNotFindLikeTableName(fullTableName);
         else
            displayCouldNotFindLikeTable(fullTableName);
      }
      return null;
   }

   /**
    * Determines if two names has the same normalized form. See
    * {@link DynamicTablesHelper#normalizeName(String)} for more information.
    *
    * @param  name1
    *         First name to compare.
    * @param  name2
    *         Second name to compare.
    *
    * @return <code>true</code> if two names has the same normalized form.
    */
   private boolean hasSameNormalizedName(String name1, String name2)
   {
      return normalizeName(name1).equals(normalizeName(name2));
   }

   /**
    * Returns a new field with the new specified name and all other parameters copied from the
    * source field.
    *
    * @param  targetName
    *         The name of the new field.
    * @param  srcField
    *         Source field to copy.
    *
    * @return see above.
    */
   private P2JField copyField(String targetName, P2JField srcField)
   {
      return new P2JField(targetName,
                          srcField.getType(),
                          srcField.getExtent(),
                          srcField.getFormat(),
                          srcField.getInitial(),
                          srcField.getLabel(),
                          srcField.getColumnLabel());

   }

  /**
   * Returns a new index with all parameters copied from the source index. Index components are
   * NOT copied.
   *
   * @param  srcIndex
   *         Source index to copy.
   *
   * @return see above.
   */
  private P2JIndex copyIndex(P2JIndex srcIndex)
   {
      return new P2JIndex(srcIndex.getTable(),
                          srcIndex.getName(),
                          srcIndex.isUnique(),
                          srcIndex.isPrimary(),
                          srcIndex.isWord());
   }

   /**
    * Add new field to the table. Doesn't check whether the field with the same name already
    * exists.
    *
    * @param field
    *        Field to add.
    */
   private void addField(P2JField field)
   {
      fields.put(normalizeName(field.getName()), field);
   }

   /**
    * Add new index to the table. Doesn't check whether the index with the same name already
    * exists.
    *
    * @param index
    *        Index to add.
    */
   private void addIndex(P2JIndex index)
   {
      indexes.put(normalizeName(index.getName()), index);
   }

   /**
    * Checks whether the field with specified name already exist.
    *
    * @param  fieldName
    *         Field name to check.
    *
    * @return <code>true</code> if there is a field which has the same name (in normalized form,
    *         see {@link DynamicTablesHelper#normalizeName(String)}).
    */
   private boolean hasField(String fieldName)
   {
      return fields.containsKey(normalizeName(fieldName));
   }

   /**
    * Checks whether the index with specified name already exist.
    *
    * @param  indexName
    *         Index name to check.
    *
    * @return <code>true</code> if there is an index which has the same name (in normalized form,
    *         see {@link DynamicTablesHelper#normalizeName(String)}).
    */
   private boolean hasIndex(String indexName)
   {
      return indexes.containsKey(normalizeName(indexName));
   }

   /**
    * Add all indexes from the given collection to the table.
    *
    * @param indexes
    *        Indexes to add.
    */
   private void addAllIndexes(Collection<P2JIndex> indexes)
   {
      if (indexes != null)
      {
         for (P2JIndex index : indexes)
         {
            addIndex(index);
         }
      }
   }

   /**
    * Add all fields from the given collection to the table.
    *
    * @param fields
    *        Fields to add.
    */
   private void addAllFields(Collection<P2JField> fields)
   {
      if (fields != null)
      {
         for (P2JField field : fields)
         {
            addField(field);
         }
      }
   }
}
