/*
** Module   : TriggerManager.java
** Abstract : trigger registration, deregistration, scoping and invocation
**
** Copyright (c) 2005-2010, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051202   @23598 Created initial version.
** 002 NVS 20060307   @24918 Fixed the trigger lookup method lookup().
** 003 NVS 20060612   @27116 Utilizing merge() method of WaitList class.
** 004 NVS 20060612   @27140 Fixed some NullPointerExceptions in trigger
**                           handling code for ANYWHERE triggers with no
**                           frame involved.
** 005 GES 20061017   @30453 Modifications to allow specific events to
**                           override the ANYWHERE processing and to be
**                           honored as triggers when they otherwise
**                           would not have been.
** 006 NVS 20070118   @31890 Reimplemented trigger registry management
**                           based on EventList class. WaitList class is
**                           no longer in use.
** 007 CA  20081204   @40819 ROW-DISPLAY event handling - these events
**                           will be deregistered from the client, when
**                           they are out of scope.
** 008 GES 20081217   @40926 Minor signature change.
** 009 GES 20090629   @42970 Remove most duplicates (except the last one) in
**                           the EventList instances when triggers are
**                           registered. This greatly reduces memory usage
**                           with no negative affect on processing since the
**                           removed instances cannot be reached nor can they
**                           be accessed via "revert".
** 010 SIY 20101122          Inferred generics. Small cleanups in javadoc.
** 011 CA  20130131          Implement ON ... PERSISTENT RUN triggers - as
**                           they are alive for the duration of the current
**                           session, when the scope which has defined them
**                           is ended, they are moved in the previous scope.
** 012 CA  20131013          Added new version of scopeFinished API, which explicitly merges a 
**                           registry into another one.
*/ 

package com.goldencode.p2j.ui;

import java.util.*;
import com.goldencode.p2j.util.*;

/**
 * This class organizes triggers registration into multiple scopes of LIFO
 * stacks.
 * <p>
 * Registrations are kept in a scoped dictionary. Every top level block creates
 * a new scope for triggers in <code>scopeStart</code> method. The registry
 * inside the scope is a <code>LinkedList</code> of <code>EventList</code>
 * objects. The list is stored as the scope extra object, so the dictionary is
 * not in use.
 * The scope's registry works as a FIFO stack for <code>EventList</code>s. The
 * most recent registration is added to the list head.
 * Deregistration may require full removal of one or more 
 * <code>EventList<code>s
 * or editing their contents. Editing happens when only some events for a
 * widget have to be deregistered, but some others remain active.
 */
public class TriggerManager
{
   /**
    * Process a notification that a new block scope is about to be entered.
    *
    * @param registry
    *        <code>ScopedDictionary</code> that keeps track of trigger
    *        registrations for this session.
    */
   public static void scopeStart(ScopedDictionary<?,?> registry)
   {
      registry.addScope(new LinkedList<EventList>());
   }
   
   /**
    * Process a notification that a scope is about to be exited.
    * <p>
    * All non-persistent triggers registered in the scope are deregistered.
    * <p>
    * All persistent triggers registered in this scope are moved to the
    * previous scope.
    *
    * @param registry
    *        <code>ScopedDictionary</code> that keeps track of trigger
    *        registrations for this session.
    */
   @SuppressWarnings("unchecked")
   public static void scopeFinished(ScopedDictionary<?,?> registry)
   {
      LinkedList<EventList> list = (LinkedList<EventList>) registry.getScope();
      registry.deleteScope();
      LinkedList<EventList> prev = (LinkedList<EventList>) registry.getScope();
      
      scopeFinished(list, prev);
   }

   /**
    * Registers a trigger within the current scope. If this is a persistent
    * trigger, all prior persistent triggers defined for the same events (on
    * any scope) will be removed.
    *
    * @param    registry
    *           Tracks trigger registrations for this session.
    * @param    events
    *           Defines the events and widgets on which to execute this trigger.
    * @param    tid
    *           The trigger ID. 
    */
   @SuppressWarnings("unchecked")
   public static void register(ScopedDictionary<?,?> registry,
                               EventList             events, 
                               int                   tid)
   {
      events.setTriggerId(tid);

      boolean persistent           = LogicalTerminal.isPersistentTrigger(tid);
      boolean match                = false;
      ArrayList<EventList>  remove = new ArrayList<EventList>();
      LinkedList<EventList> list   = null; 

      for (int i = 0; i < registry.size(); i++)
      {
         list = (LinkedList<EventList>) registry.getScopeAt(i);
         ListIterator<EventList> iter = list.listIterator(list.size());
         
         // iterate backwards through the list and remove any exact matches 
         // other than the last match in the list. for persistent triggers,
         // all matches are removed
         while (iter.hasPrevious())
         {
            EventList possible = iter.previous();
            
            if (events.isSame(possible))
            {
               int ptid = possible.getTriggerId();

               if (!match && !persistent)
               {
                  // for non-persistent triggers, if this is the first match,
                  // we leave it behind. for persistent triggers, all matches
                  // must be removed
                  match = true;
               }
               else if (persistent || 
                        !LogicalTerminal.isPersistentTrigger(ptid))
               {
                  // subsequent match is removed only in these cases:
                  // 1. a non-persistent trigger is added and this is a non-
                  //    persistent trigger
                  // 2. a persistent trigger is added
                  iter.remove();
                  remove.add(possible);
               }
            }
         }
         
         if (!persistent)
         {
            // for non-persistent trigger only the closest scope is searched
            break;
         }
      }
      
      // implement the cleanup processing for those triggers which have been 
      // removed (persistent or non-persistent)
      if (match || persistent)
      {
         Set<Integer> wids = new HashSet<Integer>(1); 
         
         for (EventList dead : remove)
         {
            cleanup(dead, wids);
         }
         
         clearRowDisplay(wids);
      }
      
      // insert the latest trigger def at the front of the list
      list = (LinkedList<EventList>) registry.getScopeAt(0);
      list.addFirst(events);
   }

   /**
    * Deregisters a trigger for specified events and widgets.
    * <p>
    * In Progress, requests to deregister (REVERT) the very first trigger
    * definition for a pair of {event, widget} are silently ignored.
    * <p>
    * Trigger deregistration is complicated. Triggers are registered for a
    * list of set pairs like:
    * <code>
    *    {(event1, event2) (widget1, widget2)}          // item 1
    *    {(event1, event2) (widget1, widget2, widget3)} // item 2
    *    ...
    * </code>
    * A deregistration request may come, for example, for event2 of widget1
    * and event1 of widget2 and event10 of widget1000. Deregistration should
    * be made for the first encountered match {event, widget} and the 
    * remaining registrations should be kept. So, in that case: 
    * <code>
    *    {(event1) (widget1)}                           // item 1.1
    *    {(event2) (widget2)}                           // item 1.2
    *    {(event1, event2) (widget1, widget2, widget3)} // item 2
    *    ...
    * </code>
    * is the result.
    *
    * @param registry
    *        <code>ScopedDictionary</code> that keeps track of trigger
    *        registrations for this session.
    * @param el
    *        <code>EventList</code> object, encoding events and widgets this
    *        request has come for
    */
   @SuppressWarnings("unchecked")
   public static void deregister(ScopedDictionary<?,?> registry, EventList el)
   {
      // First of all, the incoming request for deregistration has to be
      // "distilled" - all combinations of {event, widget} that refer to
      // the only existing registration have to be removed.
      // TO-DO implementation
      
      // Create a list of deregistration jobs and populate it with the
      // original request. The jobs list is required to keep tracks of all
      // fractions of the original event list which appear as deregistration
      // progresses.
      LinkedList<EventList> lel = new LinkedList<EventList>();
      lel.add(el);

      // loop over all scopes
      LinkedList<EventList> registryScope = null;
      
      for (int i = 0; i < registry.size(); i ++)
      {
         registryScope = (LinkedList<EventList>) registry.getScopeAt(i);
         LinkedList<EventList> m = new LinkedList<EventList>();
         boolean modified = false;

         // loop over all registry EventLists of the scope
         Iterator<EventList> iter = registryScope.iterator();
         EventList rel = null;
         
         while (iter.hasNext())
         {
            // get the next event list from registry
            rel = iter.next();
            
            // check the event against all entries of deregistration job list
            Iterator<EventList> jobs = lel.iterator();
            EventList cel = null;
            LinkedList<EventList> fractions = new LinkedList<EventList>();
            int mods = 0;

            while (jobs.hasNext())
            {
               // get the next "job" event
               cel = jobs.next();
               
               // check if unrelated events or widgets
               if (!rel.intersects(cel))
               {
                  fractions.add(cel);
                  continue;
               }
               
               // determine the deregistration and difference lists
               EventList dereg = rel.intersection(cel);
               EventList diff  = rel.difference(cel);
               
               // put the difference back to the modified registry
               m.add(diff);
               mods++;
               
               // subtract the deregistration list from the job
               EventList fracts = cel.difference(dereg);
               if (fracts != null)
                  fractions.add(fracts);
            }
            
            // save the modified jobs list
            lel = fractions;
            
            // save the modified registry
            if (mods == 0)
            {
               m.add(rel);
               continue;
            }
            
            modified = true;
         }
         
         // check if the registry was modified at all
         if (!modified)
            continue;
         
         // submit the modified list back to the registry
         registry.setScopeAt(i, m);
      }
   }

   /**
    * Merges all trigger registry entries of all scopes and produces a
    * <code>EventList</code> instance that represents all sensitive widgets
    * with all event codes they are sensitive to.
    *
    * @param   additional
    *          An additional list which needs to be merged first.
    * @param   registry
    *          <code>ScopedDictionary</code> that keeps track of trigger
    *          registrations for this session.
    *
    * @return  a merged <code>EventList</code> of all registry entries.
    */
   @SuppressWarnings("unchecked")
   public static EventList mergeTriggerRegistry(Collection<LinkedList<EventList>> additional,
                                                ScopedDictionary<?,?>             registry) 
   {
      EventList merge = new EventList();
      LinkedList<EventList> registryScope = null;
      
      // add the additional triggers first; they may be overridden by the registry triggers.
      for (LinkedList<EventList> list : additional)
      {
         for (EventList el : list)
         {
            merge.merge(el,  false);
         }
      }
      
      // loop over all scopes
      EventList el = null;
      for (int i = 0; i < registry.size(); i ++)
      {
         registryScope = (LinkedList<EventList>)registry.getScopeAt(i);

         // loop over all registry EventLists of the scope
         Iterator<EventList> iter = registryScope.iterator();
         while (iter.hasNext())
         {
            el = iter.next();
            merge.merge(el, false);
         }
      }

      return merge;
   }
   
   /**
    * Process a notification that a scope is about to be exited.
    * <p>
    * All non-persistent triggers registered in the scope are deregistered.
    * <p>
    * All persistent triggers registered in this scope are moved to the
    * previous scope.
    *
    * @param    source
    *           The source event list.
    * @param    target
    *           The target where the persistent triggers will be merged.
    */
   @SuppressWarnings("unchecked")
   static void scopeFinished(LinkedList<EventList> source, LinkedList<EventList> target)
   {
      if (source.size() == 0)
      {
         return;
      }
      
      Set<Integer> wids = new HashSet<Integer>(1);
   
      for (EventList events : source)
      {
         // all persistent triggers created in this scope will be moved on the
         // front of the list in the previous scope
         // prev can be null only when the global scope is ending
         if (target != null && 
             LogicalTerminal.isPersistentTrigger(events.getTriggerId()))
         {
            target.addFirst(events);
         }
         else
         {
            cleanup(events, wids);
         }
      }
      
      clearRowDisplay(wids);
   }

   /**
    * Handle the cleanup associated with a given event list. The associated
    * trigger will be deregistered from the LogicalTerminal's trigger
    * registry. In addition, any widgets that need notifications for row
    * display events, will be removed from the client side list.
    *
    * @param    list
    *           The event list to be deregistered.
    * @param    triggers
    *           The list of widgets no longer needing row display notifications.
    *           This will be modified on output to add any widgets that need
    *           to be removed from the notification list.
    */
   private static void cleanup(EventList list, Set<Integer> triggers)
   {
      LogicalTerminal.deregisterTrigger(list.getTriggerId(), true);

      Map<Integer, Integer> rde = list.getRowDisplayEvents();
      
      if (rde != null)
      {
         for (int wid : rde.keySet())
         {
            triggers.add(rde.get(wid));
         }
      }
   }
   
   /**
    * Remove the given widgets from the client-side list of widgets which must
    * get notifications of row display events.
    *
    * @param    triggers
    *           The list of widgets no longer needing row display notifications.
    */
   private static void clearRowDisplay(Set<Integer> triggers)
   {
      if (triggers.size() > 0)
      {
         int[] triggerIds = Utils.integerCollectionToPrimitive(triggers);
         LogicalTerminal.removeRowDisplayEvents(triggerIds);
      }
   }
}
